from abc import abstractmethod

from .compatible_abc import CompatibleABC


class AbstractOptional(CompatibleABC):
    @abstractmethod
    def is_empty(self):
        pass

    @abstractmethod
    def get(self):
        pass

    @abstractmethod
    def get_or_default(self, default_value):
        pass

    @abstractmethod
    def get_or_raise(self, raiseable):
        pass

    @abstractmethod
    def if_present(self, consumer):
        pass

    @abstractmethod
    def or_else(self, supplier):
        pass

    @abstractmethod
    def or_else_raise(self, raiseable):
        pass

    @abstractmethod
    def map(self, func):
        pass

    @abstractmethod
    def flat_map(self, func):
        pass

    def is_present(self):
        return not self.is_empty()

    __bool__ = __nonzero__ = is_present



import os
import sys
from pathlib import Path

from project_checker.DiskModelChecker import (
    _get_parent_path,
    ErrorCode,
    ErrorAndWarningData,
    _dedent_all,
)

ERROR_CODE = "error_code"
EXPECTEDS = "expecteds"

EXPECTED_INTERNAL = {
    # test case with bad xml in one of the files (contains a raw & symbol derived from ccpn internal
    "231211_hproNGF.ccpn": {
        ERROR_CODE: ErrorCode.BAD_XML,
        EXPECTEDS: [
            ErrorAndWarningData(
                code=ErrorCode.BAD_XML,
                cause="../test_data/231211_hproNGF.ccpn/ccpnv3/ccp/nmr/Nmr/prongf+prongf_user_2023-12-14-10-44-42-561_00001.xml",
                detail=_dedent_all("""\
                               could not read the file ../test_data/231211_hproNGF.ccpn/ccpnv3/ccp/nmr/Nmr/prongf+prongf_user_2023-12-14-10-44-42-561_00001.xml because\n'
                               while xml parsing ../test_data/231211_hproNGF.ccpn/ccpnv3/ccp/nmr/Nmr/prongf+prongf_user_2023-12-14-10-44-42-561_00001.xml 
                               i got the error not well-formed (invalid token): line 27381, column 381"""),
            ),
            ErrorAndWarningData(
                code=ErrorCode.EXO_LINKED_FILE_HAS_WRONG_KEY,
                cause="../test_data/231211_hproNGF.ccpn/ccpnv3/ccp/molecule/ChemElement/CCPN_reference+msd_ccpnRef_2007-12-11-11-36-46_00001.xml",
                detail=_dedent_all("""\
                               in the exo linked file CCPN_reference+msd_ccpnRef_2007-12-11-11-36-46_00001.xml
                               the key name [index 1] in the original link does not match the key expected key from the filename
                               key in the file name: CCPN reference key in the root file contents: CCPN_reference"""),
            ),
        ],
    }
}

INTERNAL_TEST_DATA_PATH = _get_parent_path(Path(os.getcwd()), 2) / "internal_test_data"
if INTERNAL_TEST_DATA_PATH.exists():
    INTERNAL_PROJECTS = INTERNAL_TEST_DATA_PATH.glob("*.ccpn")
else:
    INTERNAL_PROJECTS = []
    print(
        f"internal project test directory [{INTERNAL_TEST_DATA_PATH}] doesn't exist, tests skipped",
        file=sys.stderr,
    )



import json
import os
import stat

from pathlib import Path

from project_checker.DiskModelChecker import ModelChecker, ExitStatus
import pytest

from .api_test_data_internal import EXPECTED_INTERNAL, INTERNAL_PROJECTS, INTERNAL_TEST_DATA_PATH, ERROR_CODE, EXPECTEDS
from .api_test_data import expecteds, \
    ERROR_CODES_READ_PROTECTED, FILES_TO_PROTECT_BY_ERROR_CODE, ERROR_CODES_NOT_READ_PROTECTED
from .testing_utils import get_test_project_and_working_directory, setup_read_protected, \
    assert_error_in_expected_and_return, check_results_against_expected


@pytest.fixture
def cleanup_write_protected_files(tmp_path):
    yield
    for dirpath, dirnames, filenames in os.walk(tmp_path):
        for dirname in dirnames:
            dir_full_path = os.path.join(dirpath, dirname)
            current_permissions = stat.S_IMODE(os.lstat(dir_full_path).st_mode)
            os.chmod(dir_full_path, current_permissions | stat.S_IREAD)
        for filename in filenames:
            file_full_path = os.path.join(dirpath, filename)
            current_permissions = stat.S_IMODE(os.lstat(file_full_path).st_mode)
            os.chmod(file_full_path, current_permissions | stat.S_IREAD)

@pytest.mark.parametrize(
    "test_case",
    ERROR_CODES_READ_PROTECTED
)
@pytest.mark.usefixtures("cleanup_write_protected_files")
def test_memops_checker_read_protected(test_case, tmp_path, time_machine):
    time_machine.move_to(0, tick=False)
    file_path, expected = expecteds[test_case]
    read_protected_files = FILES_TO_PROTECT_BY_ERROR_CODE[test_case]

    tests_directory = Path(__file__).parent

    file_path = tests_directory / file_path

    file_path = setup_read_protected(file_path, tmp_path, read_protected_files)

    project_path, working_directory = get_test_project_and_working_directory(file_path)

    check_results_against_expected(expected, project_path, working_directory)



@pytest.mark.parametrize(
    "test_case",
    ERROR_CODES_NOT_READ_PROTECTED
    # ['EXO_LINKS_WITH_KEYS_OUTSIDE_THE_CCPN_CHARACTER_SET']
)
def test_memops_checker_normal(test_case, time_machine):
    time_machine.move_to(0, tick=False)
    file_path, expected = expecteds[test_case]

    tests_directory = Path(__file__).parent

    file_path = tests_directory / file_path

    project_path, working_directory = get_test_project_and_working_directory(file_path)

    check_results_against_expected(expected, project_path, working_directory)



def test_internal_error():
    file_path, expected = expecteds['GOOD_PROJECT_WITHOUT_ERRORS']
    checker = ModelChecker()

    def raise_exception(*args, **kwargs):
        raise Exception('test exception')

    checker._get_project_name = raise_exception
    result = checker.run(file_path)

    assert result == ExitStatus.EXIT_INTERNAL_ERROR


def test_Mfull_assignment_pH475_LH2():
    path = '/Users/garyt/Downloads/Mfull_assignment_pH475_LH2.ccpn'
    checker = ModelChecker()
    result = checker.run(path)

    for message in checker.messages:
        print(message[0])

    if len(checker.errors) > 0:
        print('There were errors')

@pytest.mark.parametrize(
    "file_path",
    INTERNAL_PROJECTS
)
def test_internal_tests(file_path, time_machine):
    time_machine.move_to(0, tick=False)
    checker = ModelChecker()

    result = checker.run(INTERNAL_TEST_DATA_PATH / file_path)

    if file_path in EXPECTED_INTERNAL:
        error_code = EXPECTED_INTERNAL[file_path][ERROR_CODE]
        assert result == error_code

        expected_errors = list(EXPECTED_INTERNAL[file_path][EXPECTEDS])

        for error in checker.errors:

                found_error = assert_error_in_expected_and_return(error, expected_errors)
                if found_error:
                    found_error = found_error.get()
                    expected_errors.remove(found_error)
                else:
                    print(error)
                    print('failure is', '\n'.join(found_error.messages))


        assert expected_errors == []


def test_get_top_object_key_types():

    # these are the types currently used as top object keys...

    EXPECTED_TYPES = ['Word', 'MolSystem', 'LongWord', 'NmrProject', 'MolType', 'Line', 'PositiveInt', 'Int']

    model_info_path = Path(__file__).parent.parent / 'model_info' / 'v_3_1_0_object_info.json'

    if not model_info_path.exists():
        pytest.skip(f'Model info file found, expected path was {str(model_info_path)}')

    with open(model_info_path, 'r') as fh:
        data = fh.read()

    json_data = json.loads(data)

    types = set()
    for object in json_data.values():
        name = object['name']

        role_keys = ['MolType', 'MolSystem', 'NmrProject']
        for super_object in object['supertype_names'] or name in role_keys:
            if 'TopObject' in super_object:
                for key in object['keys']:
                    key_type = object['key_types_names'][key][-1]

                    # for information
                    # print(f'{name}: {key}, {key_type}')
                    types.add(key_type)

    assert types == set(EXPECTED_TYPES)








from .abstract_optional import AbstractOptional
from .exceptions import FlatMapFunctionDoesNotReturnOptionalException


class Something(AbstractOptional):
    def __init__(self, value, optional):
        if value is None:
            raise ValueError("Invalid value for Something: None")

        self.__value = value
        self.__optional = optional
        self.messages = ()
        self.error_code = None
        self.cause = None

    def is_empty(self):
        return False

    def get(self):
        return self.__value

    def get_or_default(self, default_value):
        return self.get()

    def get_or_raise(self, raiseable):
        return self.get()

    def if_present(self, consumer):
        consumer(self.get())
        return self

    def or_else(self, supplier):
        return self

    def or_else_raise(self, raiseable):
        return self

    def map(self, func):
        return self.__optional.of(func(self.get()))

    def flat_map(self, func):
        res = func(self.get())
        if not isinstance(res, AbstractOptional):
            raise FlatMapFunctionDoesNotReturnOptionalException(
                "Mapping function to flat_map must return Optional."
            )

        return res

    def __eq__(self, other):
        return isinstance(other, Something) and self.get() == other.get()

    def __repr__(self):
        return "Optional.of({!r})".format(self.get())



from pathlib import Path

from project_checker.DiskModelChecker import ErrorAndWarningData, ErrorCode, _dedent_all


BAD_EXO_3_CONTENT_CORRECTIONS = {
    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00004 MOLS.MolSystem - xml is bad skipped [see errors at the end of the run for details]",
    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                  [see complete errors at the end of the the run for details]"""),
}
EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT = [
    # tranche 0 always runs
    ("analysis took 0.000 seconds", False),
    # tranche 1 pre good memops root
    ("target {project_name}.ccpn", False),
    ("project_name appears to be... {project_name}", False),
    ("the directory {project_name}.ccpn has the correct suffix", False),
    (
        "found an implementation directory {project_name}.ccpn/ccpnv3/memops/Implementation",
        False,
    ),
    (
        "the path {project_name}.ccpn/ccpnv3/memops/Implementation/{project_name}.xml is a possible memops root [name matches project]",
        False,
    ),
    ("The project in {project_name}.xml, was not renamed after saving", False),
    # tranche 2 post good memops root
    (
        "ccpn project memops root file found in {project_name}.ccpn/ccpnv3/memops/Implementation/{project_name}.xml",
        False,
    ),
    ("model version that saved this file appears to be 3.1.0", False),
    ("memops root data was stored at Sat Feb 24 16:16:06 2024", False),
    ("ccpnmr program version that saved this file appears to be 3.2.1", False),
]


EXPECTED_GOOD_EXO_LINKS = [
    ("searching for top object exo links, found 8", False),
    ("analysing exo links", False),
    (
        "  1. default_user_2024-02-24-15-54-35-583_00006 GUIT.GuiTask [keys: {{'nameSpace': 'user', 'name': 'View'}}]",
        True,
    ),
    (
        "  2. IsoSchemeProj_user_2008-08-01-11-46-16_00022 CCLB.LabelingScheme [keys: {{'name': 'uni_15N'}}]",
        True,
    ),
    (
        "  3. default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem [keys: {{'code': 'default'}}]",
        True,
    ),
    (
        "  4. cam_wb104_2008-01-15-16-06-39_00031 NMRX.NmrExpPrototype [keys: {{'serial': '32'}}]",
        True,
    ),
    (
        "  5. default_user_2024-02-24-15-54-35-583_00001 NMR.NmrProject [keys: {{'name': 'default'}}]",
        True,
    ),
    (
        "  6. default_user_2024-02-24-15-54-35-583_00003 REFS.RefSampleComponentStore [keys: {{'name': 'default'}}]",
        True,
    ),
    (
        "  7. default_user_2024-02-24-15-54-35-583_00002 SAM.SampleStore [keys: {{'name': 'default'}}]",
        True,
    ),
    (
        "  8. default_user_2024-02-24-15-54-35-583_00005 GUIW.WindowStore [keys: {{'nmrProject': '_ccp_nmr_Nmr_NmrProject___default___'}}]",
        True,
    ),
    ("found 8 out of 8 top object files exo linked by the project", False),
    ("expected top object paths are:", False),
    ("  1. default_user_2024-02-24-15-54-35-583_00006 GUIT.GuiTask - [PROJECT] ccpnmr/gui/Task/user+View+default_user_2024-02-24-15-54-35-583_00006.xml", True),
    ("  2. IsoSchemeProj_user_2008-08-01-11-46-16_00022 CCLB.LabelingScheme - [REFERENCE] ccp/molecule/ChemCompLabel/uni_15N+IsoSchemeProj_user_2008-08-01-11-46-16_00022.xml", True),
    ("  3. default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem - [PROJECT] ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml", True),
    ("  4. cam_wb104_2008-01-15-16-06-39_00031 NMRX.NmrExpPrototype - [REFERENCE] ccp/nmr/NmrExpPrototype/32+cam_wb104_2008-01-15-16-06-39_00031.xml", True),
    ("  5. default_user_2024-02-24-15-54-35-583_00001 NMR.NmrProject - [PROJECT] ccp/nmr/Nmr/default+default_user_2024-02-24-15-54-35-583_00001.xml", True),
    ("  6. default_user_2024-02-24-15-54-35-583_00003 REFS.RefSampleComponentStore - [PROJECT] ccp/lims/RefSampleComponent/default+default_user_2024-02-24-15-54-35-583_00003.xml", True),
    ("  7. default_user_2024-02-24-15-54-35-583_00002 SAM.SampleStore - [PROJECT] ccp/lims/Sample/default+default_user_2024-02-24-15-54-35-583_00002.xml", True),
    ("  8. default_user_2024-02-24-15-54-35-583_00005 GUIW.WindowStore - [PROJECT] ccpnmr/gui/Window/_ccp_nmr_Nmr_NmrProject___default___+default_user_2024-02-24-15-54-35-583_00005.xml", True),

    ("checking the contents of 8 linked top objects", False),
    (
        "  1. default_user_2024-02-24-15-54-35-583_00006 GUIT.GuiTask - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
        True,
    ),
    (
        "  2. IsoSchemeProj_user_2008-08-01-11-46-16_00022 CCLB.LabelingScheme - is ok [reference object assumed good (further analysis skipped)]",
        True,
    ),
    (
        "  3. default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
        True,
    ),
    (
        "  4. cam_wb104_2008-01-15-16-06-39_00031 NMRX.NmrExpPrototype - is ok [reference object assumed good (further analysis skipped)]",
        True,
    ),
    (
        "  5. default_user_2024-02-24-15-54-35-583_00001 NMR.NmrProject - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
        True,
    ),
    (
        "  6. default_user_2024-02-24-15-54-35-583_00003 REFS.RefSampleComponentStore - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
        True,
    ),
    (
        "  7. default_user_2024-02-24-15-54-35-583_00002 SAM.SampleStore - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
        True,
    ),
    (
        "  8. default_user_2024-02-24-15-54-35-583_00005 GUIW.WindowStore - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
        True,
    ),
    (
        "all the analysed linked top objects [8] appear to have the correct basic structure",
        False,
    ),
    ("checking the exo link keys in 8 top object file names", False),
    (
        "  1. default_user_2024-02-24-15-54-35-583_00006 GUIT.GuiTask - all keys are good",
        True,
    ),
    (
        "  2. IsoSchemeProj_user_2008-08-01-11-46-16_00022 CCLB.LabelingScheme - all keys are good",
        True,
    ),
    (
        "  3. default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem - all keys are good",
        True,
    ),
    (
        "  4. cam_wb104_2008-01-15-16-06-39_00031 NMRX.NmrExpPrototype - all keys are good",
        True,
    ),
    (
        "  5. default_user_2024-02-24-15-54-35-583_00001 NMR.NmrProject - all keys are good",
        True,
    ),
    (
        "  6. default_user_2024-02-24-15-54-35-583_00003 REFS.RefSampleComponentStore - all keys are good",
        True,
    ),
    (
        "  7. default_user_2024-02-24-15-54-35-583_00002 SAM.SampleStore - all keys are good",
        True,
    ),
    (
        "  8. default_user_2024-02-24-15-54-35-583_00005 GUIW.WindowStore - all keys are good",
        True,
    ),
    ("8 of the 8 keys are good", False),
]


def update_notes(
    templates,
    file_name="",
    sub_directory="",
    replace_text={},
    replace_item={},
    delete=(),
    discard_beyond=None,
):
    def expand_file_name_template(template, project_name, sub_directory):
        return eval(f'f"""{template}"""')

    templates = list(templates)

    # TODO move this after template expansion
    if discard_beyond:
        new_templates = []
        for item in templates:
            if discard_beyond in item[0]:
                new_templates.append(item)
                break
            else:
                new_templates.append(item)
        templates = new_templates

    for target, replacement in replace_text.items():
        for i, template in enumerate(templates):
            if target in template[0]:
                templates[i] = (template[0].replace(target, replacement), template[1])

    sub_directory = f"/{sub_directory}" if sub_directory else ""
    result = [
        (
            expand_file_name_template(note[0], file_name, sub_directory),
            note[1],
        )
        for note in templates
    ]

    for i, item in enumerate(result):
        if item in replace_item.items():
            result[i] = replace_item[item]

    for i, item in enumerate(result):
        for deletion_item in delete:
            if deletion_item in item:
                result[i] = None
    result = [item for item in result if item is not None]

    return result


expecteds = {
    "MISSING_PROJECT": [
        "../test_data/no_such_file.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MISSING_DIRECTORY,
                Path("no_such_file.ccpn"),
                "the directory no_such_file.ccpn doesn't exist",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "no_such_file",
                discard_beyond="project_name appears to be",
            ),
        ],
    ],
    "PROJECT_IS_FILE": [
        "../test_data/project_is_file.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.IS_NOT_DIRECTORY,
                Path("project_is_file.ccpn"),
                "the path project_is_file.ccpn should be a directory, it isn't, its a file",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "project_is_file",
                discard_beyond="project_name appears to be",
            ),
        ],
    ],
    "EMPTY_PROJECT": [
        "../test_data/empty_project.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MISSING_DIRECTORY,
                Path("empty_project.ccpn/ccpnv3"),
                "the directory empty_project.ccpn/ccpnv3 doesn't exist",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "empty_project",
                discard_beyond="has the correct suffix",
            ),
        ],
    ],
    "EMPTY_PROJECT_NO_EXTENSION": [
        "../test_data/empty_project_no_extension_ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MISSING_DIRECTORY,
                Path("empty_project_no_extension_ccpn/ccpnv3"),
                "the directory empty_project_no_extension_ccpn/ccpnv3 doesn't exist",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "empty_project_no_extension_ccpn",
                discard_beyond="project_name appears to be",
                replace_text={".ccpn": ""},
            ),
            (
                "ccpn project directories should have the suffix .ccpn the directory empty_project_no_extension_ccpn doesn't",
                False,
            ),
        ],
    ],
    "READ_PROTECTED_PROJECT": [
        "../test_data/no_read_empty_project.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NOT_READABLE,
                Path("no_read_empty_project.ccpn"),
                "the path no_read_empty_project.ccpn is not readable",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "no_read_empty_project",
                discard_beyond="project_name appears to be",
            ),
        ],
    ],
    "EMPTY_CCPNV3": [
        "../test_data/empty_ccpnv3.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MISSING_DIRECTORY,
                Path("empty_ccpnv3.ccpn/ccpnv3/memops/Implementation"),
                "the directory empty_ccpnv3.ccpn/ccpnv3/memops/Implementation doesn't exist",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "empty_ccpnv3",
                discard_beyond=" has the correct suffix",
            ),
        ],
    ],
    "READ_PROTECTED_CCPNV3": [
        "../test_data/no_read_ccpnv3.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NOT_READABLE,
                Path("no_read_ccpnv3.ccpn/ccpnv3"),
                "the path no_read_ccpnv3.ccpn/ccpnv3 is not readable",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "no_read_ccpnv3",
                discard_beyond="has the correct suffix",
            ),
        ],
    ],
    "EMPTY_IMPLEMENTATION": [
        "../test_data/empty_implementation.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NO_MEMOPS_ROOT_FILES,
                Path("empty_implementation.ccpn/ccpnv3/memops/Implementation"),
                """no xml files which could be memops roots were found in empty_implementation.ccpn/ccpnv3/memops/Implementation
                         with the following extra messages being reported no root xml file found in empty_implementation.ccpn/ccpnv3/memops/Implementation""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "empty_implementation",
                discard_beyond="found an implementation directory",
            ),
        ],
    ],
    "MEMOPS_ROOT_XML_IS_DIRECTORY": [
        "../test_data/memops_root_xml_is_directory.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.IS_NOT_FILE,
                Path(
                    "memops_root_xml_is_directory.ccpn/ccpnv3/memops/Implementation/memops_root_xml_is_directory.xml"
                ),
                "memops_root_xml_is_directory.ccpn/ccpnv3/memops/Implementation/memops_root_xml_is_directory.xml is not a file it's a directory",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "memops_root_xml_is_directory",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "GOOD_PROJECT_WITHOUT_ERRORS": [
        "../test_data/good_projects/empty_good_project.ccpn",
        [
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project",
                sub_directory="good_projects",
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
        ],
    ],
    "GOOD_PROJECT_WITHOUT_ERRORS_RENAMED": [
        "../test_data/good_projects/empty_good_project_renamed.ccpn",
        [
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project_renamed",
                sub_directory="good_projects",
                replace_text={
                    "was probably renamed after saving": "was not renamed after saving"
                },
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
        ],
    ],
    "MEMOPS_ROOT_READ_PROTECTED_GOOD_PROJECT": [
        "../test_data/memops_root_no_read_empty_good_project.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NOT_READABLE,
                Path(
                    "memops_root_no_read_empty_good_project.ccpn/ccpnv3/memops/Implementation/memops_root_no_read_empty_good_project.xml"
                ),
                """no xml files which could be memops roots were found in memops_root_no_read_empty_good_project.ccpn/ccpnv3/memops/Implementation
                          with the following extra messages being reported the file memops_root_no_read_empty_good_project.ccpn/ccpnv3/memops/Implementation/memops_root_no_read_empty_good_project.xml is not a valid root xml file and
                          while passing this file i got the following messages:
                          while reading memops_root_no_read_empty_good_project.ccpn/ccpnv3/memops/Implementation/memops_root_no_read_empty_good_project.xml i got the error [Errno 13] Permission denied: 'memops_root_no_read_empty_good_project.ccpn/ccpnv3/memops/Implementation/memops_root_no_read_empty_good_project.xml
                      """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                "memops_root_no_read_empty_good_project",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "BAD_MEMOPS_ROOT_XML": [
        "../test_data/bad_memops_root_xml.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BAD_XML,
                Path(
                    "bad_memops_root_xml.ccpn/ccpnv3/memops/Implementation/bad_memops_root_xml.xml"
                ),
                """no xml files which could be memops roots were found in bad_memops_root_xml.ccpn/ccpnv3/memops/Implementation
                   with the following extra messages being reported the file bad_memops_root_xml.ccpn/ccpnv3/memops/Implementation/bad_memops_root_xml.xml is not a valid root xml file and
                   while passing this file i got the following messages:
                   while xml parsing bad_memops_root_xml.ccpn/ccpnv3/memops/Implementation/bad_memops_root_xml.xml i got the error attributes construct error, line 1, column 36 (<string>, line 1)'
                """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="bad_memops_root_xml",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "MEMOPS_ROOT_NO_STORAGE_UNIT": [
        "../test_data/memops_root_no_storage_unit.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NO_STORAGE_UNIT,
                Path(
                    "memops_root_no_storage_unit.ccpn/ccpnv3/memops/Implementation/memops_root_no_storage_unit.xml"
                ),
                """no xml files which could be memops roots were found in memops_root_no_storage_unit.ccpn/ccpnv3/memops/Implementation
                   with the following extra messages being reported the file memops_root_no_storage_unit.ccpn/ccpnv3/memops/Implementation/memops_root_no_storage_unit.xml is not a valid root xml file and
                   while passing this file i got the following messages:
                   expected storage unit but found NO-StorageUnit at root of memops_root_no_storage_unit.ccpn/ccpnv3/memops/Implementation/memops_root_no_storage_unit.xml""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="memops_root_no_storage_unit",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "NO_MEMOPS_ROOT_ELEMENT": [
        "../test_data/empty_storage_unit.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NO_ROOT_OR_TOP_OBJECT,
                Path(
                    "empty_storage_unit.ccpn/ccpnv3/memops/Implementation/empty_storage_unit.xml"
                ),
                """no xml files which could be memops roots were found in empty_storage_unit.ccpn/ccpnv3/memops/Implementation
                with the following extra messages being reported
                the file empty_storage_unit.ccpn/ccpnv3/memops/Implementation/empty_storage_unit.xml is not a valid root xml file and
                while passing this file i got the following messages:
                expected single root element under storage unit, found none
                """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_storage_unit",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "ROOT_NOT_MEMOPS_ROOT_ELEMENT": [
        "../test_data/root_not_memops_root_element.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.ROOT_IS_NOT_MEMOPS_ROOT,
                Path(
                    "root_not_memops_root_element.ccpn/ccpnv3/memops/Implementation/root_not_memops_root_element.xml"
                ),
                """no xml files which could be memops roots were found in root_not_memops_root_element.ccpn/ccpnv3/memops/Implementation
                      with the following extra messages being reported
                      the file root_not_memops_root_element.ccpn/ccpnv3/memops/Implementation/root_not_memops_root_element.xml is not a valid root xml file and
                      while passing this file i got the following messages:
                      root_not_memops_root_element.ccpn/ccpnv3/memops/Implementation/root_not_memops_root_element.xml doesn't contain a IMPL.MemopsRoot
                      """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_not_memops_root_element",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "TOO_MANY_ROOT_ELEMENTS": [
        "../test_data/too_many_root_elements.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MULTIPLE_ROOT_OR_TOP_OBJECTS_IN_STORAGE,
                Path(
                    "too_many_root_elements.ccpn/ccpnv3/memops/Implementation/too_many_root_elements.xml"
                ),
                """'no xml files which could be memops roots were found in too_many_root_elements.ccpn/ccpnv3/memops/Implementation
                      with the following extra messages being reported the file too_many_root_elements.ccpn/ccpnv3/memops/Implementation/too_many_root_elements.xml is not a valid root xml file and
                      while passing this file i got the following messages:
                      expected single root element under storage unit, found 2
                      """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="too_many_root_elements",
                discard_beyond="is a possible memops root",
            ),
        ],
    ],
    "NO_EXO_LINKS": [
        "../test_data/no_exo_links.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NO_EXO_LINKS_FOUND,
                Path("no_exo_links.ccpn/ccpnv3/memops/Implementation/no_exo_links.xml"),
                """no exo links were found in the memops root file no_exo_links.ccpn/ccpnv3/memops/Implementation/no_exo_links.xml""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT, file_name="no_exo_links"
            ),
            ("analysing exo links", False),
            ("searching for top object exo links, found 0", False),
        ],
    ],
    # 'BASIC_EXO_LINKS_NOT_FOUND': [
    #     "../test_data/basic_exo_links_not_found.ccpn",
    #     [
    #         ErrorData(ErrorCode.BASIC_EXO_LINKS_NOT_FOUND,
    #                   Path('basic_exo_links_not_found.ccpn/ccpnv3/memops/Implementation/basic_exo_links_not_found.xml'),
    #
    #                   ''''in the file memops root file  ../test_data/basic_exo_links_not_found.ccpn/ccpnv3/memops/Implementation/basic_exo_links_not_found.xml
    #                       the following 7 basic exo links are missing:
    #                       GuiTask
    #                       LabelingScheme
    #                       MolSystem
    #                       NmrExpPrototype
    #                       RefSampleComponentStore
    #                       SampleStore
    #                       WindowStore'''
    #                   ),
    #         *update_notes(EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT, file_name='basic_exo_links_not_found'),
    #         ('searching for top object exo links, found 1', False),
    #         ('1. default_user_2024-02-24-15-54-35-583_00001 [type: NmrProject]', True),
    #     ]
    # ],
    "EXO_LINKED_FILE_MISSING": [
        "../test_data/delayed_errors/exo_linked_file_missing.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_LINKED_FILE_MISSING,
                "default_user_2024-02-24-15-54-35-583_00004",
                """missing top object file for exo link guid default_user_2024-02-24-15-54-35-583_00004""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_linked_file_missing",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "contents of 8": "contents of 7",
                    "found 8 out of 8": "found 7 out of 8",
                    "8 of the 8": "7 of the 7",
                    "analysed top objects [8]": "analysed top objects [7]",
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00004 MOLS.MolSystem - the file is missing",
                    "analysed linked top objects [8] appear": "analysed linked top objects [7] appear",
                    "checking the exo link keys in 8": "checking the exo link keys in 7",
                    "583_00004 MOLS.MolSystem - all keys are good": "583_00004 MOLS.MolSystem - the file is missing",
                    "ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml":
                    "*file not found*"
                },
                delete=("all the analysed top objects [8]",),
            ),
            (
                "there are 1 missing top object files the list of exo links for the missing files are:",
                False,
            ),
            (
                "1. default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem [keys: {'code': 'default'}]",
                True,
            ),
            (
                "empty directories [1] which may be orphaned containers found and listed below [warning]",
                False,
            ),
            (
                "  1. exo_linked_file_missing.ccpn/ccpnv3/ccp/molecule/MolSystem [warning]",
                True,
            ),
        ],
    ],
    "EXO_LINKED_FILE_NOT_VALID_XML": [
        "../test_data/delayed_errors/exo_linked_file_bad_xml.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BAD_XML,
                "exo_linked_file_bad_xml.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """could not read the file exo_linked_file_bad_xml.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml because
                       while xml parsing exo_linked_file_bad_xml.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml i got the error attributes construct error, line 1, column 36 (<string>, line 1) """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_linked_file_bad_xml",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS, replace_text=BAD_EXO_3_CONTENT_CORRECTIONS
            ),
        ],
    ],
    "EXO_LINKED_FILE_NO_STORAGE_UNIT": [
        "../test_data/delayed_errors/exo_linked_file_no_storage_unit.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NO_STORAGE_UNIT,
                "exo_linked_file_no_storage_unit.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """could not read the file exo_linked_file_no_storage_unit.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml because
                         expected storage unit but found _NoStorageUnit at root of exo_linked_file_no_storage_unit.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml
                      """,
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_linked_file_no_storage_unit",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS, replace_text=BAD_EXO_3_CONTENT_CORRECTIONS
            ),
        ],
    ],
    "EXO_LINKED_FILE_NO_TOP_OBJECT": [
        "../test_data/delayed_errors/exo_link_no_top_object.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NO_ROOT_OR_TOP_OBJECT,
                "exo_link_no_top_object.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """could not read the file exo_link_no_top_object.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml because
                         expected single root element under storage unit, found none""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_link_no_top_object",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS, replace_text=BAD_EXO_3_CONTENT_CORRECTIONS
            ),
        ],
    ],
    "EXO_LINKED_FILE_TOO_MANY_TOP_OBJECTS": [
        "../test_data/delayed_errors/exo_link_too_many_top_objects.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MULTIPLE_ROOT_OR_TOP_OBJECTS_IN_STORAGE,
                "exo_link_too_many_top_objects.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """could not read the file exo_link_too_many_top_objects.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml because
                         expected single root element under storage unit, found 2""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_link_too_many_top_objects",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS, replace_text=BAD_EXO_3_CONTENT_CORRECTIONS
            ),
        ],
    ],
    "INTERNAL_AND_EXTERNAL_GUIDS_DISAGREE": [
        "../test_data/delayed_errors/external_guid_doesnt_match_internal.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.INTERNAL_AND_EXTERNAL_GUIDS_DISAGREE,
                "external_guid_doesnt_match_internal.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """the guid in the file default+default_user_2024-02-24-15-54-35-583_00004.xml does not match the guid in the file name
                         file name guid: default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem
                         file guid: dummy-internal-guid-for-testing unknown""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="external_guid_doesnt_match_internal",
                sub_directory="delayed_errors",
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
        ],
    ],
    "GOOD_PROJECT_EXTRA_XML_FILES": [
        "../test_data/warn_projects/empty_good_project_extra_xml_files.ccpn",
        [
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project_extra_xml_files",
                sub_directory="warn_projects",
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
            (
                "there are 2 files in the project directory that are not linked to a file by an exo link [warning]",
                False,
            ),
            ("  1. ccp/molecule/MolSystem/molecule2.xml [warning]", True),
            ("  2. ccpnmr/gui/Task/task_2.xml [warning]", True),
            ("checking the contents of 2 detached top objects", False),
            (
                "  1. molecule2 *unknown-package*.*unknown-class* - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
                True,
            ),
            (
                "  2. task_2 *unknown-package*.*unknown-class* - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
                True,
            ),
            (
                "all the analysed detached top objects [2] appear to have the correct basic structure",
                False,
            ),
        ],
    ],
    "EXO_FILE_NO_PACKAGE_GUID": [
        "../test_data/delayed_errors/exo_file_no_package_guid.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.MISSING_PACKAGE_GUID,
                "exo_file_no_package_guid.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                "the top object has no packageGuid attribute in the element _StorageUnit in the file default+default_user_2024-02-24-15-54-35-583_00004.xml  MOLS.MolSystem",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_no_package_guid",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00004 MOLS.MolSystem - the top object has no packageGuid attribute in the element _StorageUnit in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem [ERROR]",
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                           [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_FILE_UNKNOWN_PACKAGE_GUID": [
        "../test_data/delayed_errors/exo_file_unknown_package_guid.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.UNKNOWN_PACKAGE_GUID,
                "exo_file_unknown_package_guid.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                "the package guid www.ccpn.ac.uk_Thompson_2006-08-16-14:22:54_00022 in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem is not recognised",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_unknown_package_guid",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00004 MOLS.MolSystem - the package guid www.ccpn.ac.uk_Thompson_2006-08-16-14:22:54_00022 in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem is not recognised [ERROR]",
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                           [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "BAD_ROOT_ELEMENT_NAME": [
        "../test_data/delayed_errors/exo_file_bad_root_element_name.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BAD_ROOT_ELEMENT_NAME,
                "exo_file_bad_root_element_name.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """the root elements name in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem doesn't have the correct name format
                         it should be of the form <SHORT_PACKAGE_NAME>.<TOP-OBJECT-NAME> but was MOLS-MolSystem""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_bad_root_element_name",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": _dedent_all("""583_00004 MOLS.MolSystem - the root elements name in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem doesn't have the correct name format
                               it should be of the form <SHORT_PACKAGE_NAME>.<TOP-OBJECT-NAME> but was MOLS-MolSystem"""),
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                              [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "UNKNOWN_SHORT_PACKAGE_NAME": [
        "../test_data/delayed_errors/exo_file_unknown_short_package_name.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.UNKNOWN_SHORT_PACKAGE_NAME,
                "exo_file_unknown_short_package_name.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                """the short package name (GARYT) in the root element of the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem is not recognised""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_unknown_short_package_name",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00004 MOLS.MolSystem - the short package name (GARYT) in the root element of the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem is not recognised [ERROR]",
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "SHORT_NAME_GUID_DOESNT_MATCH_PACKAGE_GUID": [
        "../test_data/delayed_errors/exo_file_short_name_guid_doesnt_match_package_guid.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.SHORT_NAME_GUID_DOESNT_MATCH_PACKAGE_GUID,
                "exo_file_short_name_guid_doesnt_match_package_guid.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                _dedent_all("""the guid for the short name CHEM [www.ccpn.ac.uk_Fogh_2006-08-16-14:22:51_00046]
                                    is not the same as the package guid www.ccpn.ac.uk_Fogh_2006-08-16-14:22:54_00022 [MOLS]
                                    for the root element in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_short_name_guid_doesnt_match_package_guid",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": _dedent_all("""583_00004 MOLS.MolSystem - the guid for the short name CHEM [www.ccpn.ac.uk_Fogh_2006-08-16-14:22:51_00046]
                              is not the same as the package guid www.ccpn.ac.uk_Fogh_2006-08-16-14:22:54_00022 [MOLS]
                              for the root element in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem [ERROR]"""),
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_FILE_TIME_ATTRIB_MISSING": [
        "../test_data/delayed_errors/exo_file_time_attrib_missing.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_FILE_TIME_ATTRIB_MISSING,
                "exo_file_time_attrib_missing.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                "in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute time is missing in the element _StorageUnit",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_time_attrib_missing",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": _dedent_all(
                        """583_00004 MOLS.MolSystem - in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute time is missing in the element _StorageUnit [ERROR]"""
                    ),
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_FILE_RELEASE_ATTRIB_MISSING": [
        "../test_data/delayed_errors/exo_file_release_attrib_missing.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_FILE_RELEASE_ATTRIB_MISSING,
                "exo_file_release_attrib_missing.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                "in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute release is missing in the element _StorageUnit",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_release_attrib_missing",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": _dedent_all(
                        """583_00004 MOLS.MolSystem - in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute release is missing in the element _StorageUnit [ERROR]"""
                    ),
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_FILE_RELEASE_DOESNT_MATCH_ROOT": [
        "../test_data/delayed_errors/exo_file_release_doesnt_match_root.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_FILE_RELEASE_DOESNT_MATCH_ROOT,
                "exo_file_release_doesnt_match_root.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                _dedent_all("""in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the model version for the top object
                                    default_user_2024-02-24-15-54-35-583_00004 [0.1.0] is different from root 3.1.0"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_release_doesnt_match_root",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": _dedent_all("""583_00004 MOLS.MolSystem - in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the model version for the top object
                                          default_user_2024-02-24-15-54-35-583_00004 [0.1.0] is different from root 3.1.0 [ERROR]"""),
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_FILE_WRONG_STORAGE_LOCATION": [
        "../test_data/delayed_errors/exo_file_wrong_storage_location.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_FILE_WRONG_STORAGE_LOCATION,
                "exo_file_wrong_storage_location.ccpn/ccpnv3/ccp/nmr/Nmr/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                _dedent_all("""the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem is not stored in the correct place in the project
                                    it should be store in ccpnv3/ccp/molecule/MolSystem but is stored in ccpnv3/ccp/nmr/Nmr"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_wrong_storage_location",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]":
                    _dedent_all("""583_00004 MOLS.MolSystem - the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem is not stored in the correct place in the project
                                   it should be stored in ccpnv3/ccp/molecule/MolSystem but is stored in ccpnv3/ccp/nmr/Nmr [ERROR]"""),
                    "all the analysed linked top objects [8] appear to have the correct basic structure":
                    _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                                   [see complete errors at the end of the the run for details]"""),
                    "MOLS.MolSystem - [PROJECT] ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml":
                    "MOLS.MolSystem - [PROJECT] ccp/nmr/Nmr/default+default_user_2024-02-24-15-54-35-583_00004.xml"
                },
            ),
        ],
    ],
    "EXO_FILE_TIME_ATTRIB_INVALID": [
        "../test_data/delayed_errors/exo_file_time_attrib_invalid.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_FILE_TIME_ATTRIB_INVALID,
                "exo_file_time_attrib_invalid.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                "in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute time [wibble] in the element _StorageUnit is not a valid time",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_time_attrib_invalid",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00004 MOLS.MolSystem - in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute time [wibble] in the element _StorageUnit is not a valid time [ERROR]",
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_FILE_RELEASE_ATTRIB_INVALID": [
        "../test_data/delayed_errors/exo_file_release_attrib_invalid.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_FILE_RELEASE_ATTRIB_INVALID,
                "exo_file_release_attrib_invalid.ccpn/ccpnv3/ccp/molecule/MolSystem/default+default_user_2024-02-24-15-54-35-583_00004.xml",
                _dedent_all("""in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute release [3.1a.0]
                                    the attribute release [3.1a.0] is not a valid version number
                                    it should be of the form <major>.<minor>.<patch>
                                    where <major>, <minor>, and <patch> can only contain digits with a possible prepended letter"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_release_attrib_invalid",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": _dedent_all("""583_00004 MOLS.MolSystem - in the file default+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem the attribute release [3.1a.0]
                                          the attribute release [3.1a.0] is not a valid version number
                                          it should be of the form <major>.<minor>.<patch>
                                          where <major>, <minor>, and <patch> can only contain digits with a possible prepended letter [ERROR]"""),
                    "all the analysed linked top objects [8] appear to have the correct basic structure": _dedent_all("""only 7 of the 8 analysed top objects appear to have the correct basic structure
                            [see complete errors at the end of the the run for details]"""),
                },
            ),
        ],
    ],
    "EXO_LINKED_FILE_HAS_WRONG_KEY": [
        "../test_data/delayed_errors/exo_file_key_doesnt_match_root.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_LINKED_FILE_HAS_WRONG_KEY,
                "exo_file_key_doesnt_match_root.ccpn/ccpnv3/ccp/molecule/MolSystem/wibble+default_user_2024-02-24-15-54-35-583_00004.xml",
                _dedent_all("""in the exo linked file wibble+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem
                                    the key code [index 1] in the original link does not match the key expected key from the filename
                                    key in the file name: default key in the root file contents: wibble
                      """),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_file_key_doesnt_match_root",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "3. default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem - all keys are good": _dedent_all("""3. [key 1] default_user_2024-02-24-15-54-35-583_00004 MOLS.MolSystem - in the exo linked file wibble+default_user_2024-02-24-15-54-35-583_00004.xml MOLS.MolSystem
                                  the key code [index 1] in the original link does not match the key expected key from the filename
                                  key in the file name: default, key in the root file contents: wibble [ERROR]"""),
                    "8 of the 8 keys are good": "7 of the 8 keys are good",
                    "MolSystem/default+default":
                    "MolSystem/wibble+default"

                },
            ),
        ],
    ],
    "BADLY_FORMATTED_ROOT_EXO_LINK": [
        "../test_data/delayed_errors/bad_root_exo_link_element.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BADLY_FORMATTED_ROOT_EXO_LINK,
                "bad_root_exo_link_element.ccpn/ccpnv3/memops/Implementation/bad_root_exo_link_element.xml",
                "the guid default_user_2024-02-24-15-54-35-583_00006 in the file bad_root_exo_link_element.xml appears to contain a badly formatted exo link",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="bad_root_exo_link_element",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "35-583_00006 GUIT.GuiTask [keys: {{'nameSpace': 'user', 'name': 'View'}}]": "35-583_00006 *unknown-package*.*unknown-class* - exo link detected but is incorrectly defined in root [error]",
                    "583_00006 GUIT.GuiTask - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]": "583_00006 *unknown-package*.*unknown-class* - is ok [saved on: Sat Feb 24 16:16:06 2024 model version: 3.1.0]",
                    "583_00006 GUIT.GuiTask - all keys are good": "583_00006 *unknown-package*.*unknown-class* - all keys are good",
                    "GUIT.GuiTask":
                    "None.None"
                },
            ),
            (
                "Error: the guid default_user_2024-02-24-15-54-35-583_00006 in the file bad_root_exo_link_element.xml appears to contain a badly formatted exo link",
                False,
            ),
        ],
    ],
    "BADLY_FORMATTED_EXO_LINK_0_CHILD_ELEMENTS": [
        "../test_data/delayed_errors/bad_embedded_exo_link_0_child_elements.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BADLY_FORMATTED_EXO_LINK_KEY_DATA,
                "bad_embedded_exo_link_0_child_elements.ccpn/ccpnv3/memops/Implementation/bad_embedded_exo_link_0_child_elements.xml",
                _dedent_all("""in the file bad_embedded_exo_link_0_child_elements.xml the link REFS.RefSampleComponentStore with guid default_user_2024-02-24-15-54-35-583_00003
                                    has the wrong number of children for key name [0]"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="bad_embedded_exo_link_0_child_elements",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00003 REFS.RefSampleComponentStore [keys: {{'name': 'default'}}]": "583_00003 REFS.RefSampleComponentStore - exo link detected but is incorrectly defined in root [error]"
                },
            ),
            (
                _dedent_all("""in the file bad_embedded_exo_link_0_child_elements.xml the link REFS.RefSampleComponentStore with guid default_user_2024-02-24-15-54-35-583_00003
                           has the wrong number of children for key name [0]"""),
                False,
            ),
        ],
    ],
    "BADLY_FORMATTED_EXO_LINK_0_ELEMENTS": [
        "../test_data/delayed_errors/bad_embedded_exo_link_0_elements.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BADLY_FORMATTED_EXO_LINK_KEY_DATA,
                "bad_embedded_exo_link_0_elements.ccpn/ccpnv3/memops/Implementation/bad_embedded_exo_link_0_elements.xml",
                _dedent_all("""in the file bad_embedded_exo_link_0_elements.xml in the link REFS.RefSampleComponentStore with guid default_user_2024-02-24-15-54-35-583_00003
                           the key name has the wrong number of children (0) in its main element"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="bad_embedded_exo_link_0_elements",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00003 REFS.RefSampleComponentStore [keys: {{'name': 'default'}}]": "583_00003 REFS.RefSampleComponentStore - exo link detected but is incorrectly defined in root [error]"
                },
            ),
            (
                _dedent_all("""in the file bad_embedded_exo_link_0_elements.xml in the link REFS.RefSampleComponentStore with guid default_user_2024-02-24-15-54-35-583_00003
                           the key name has the wrong number of children (0) in its main element"""),
                False,
            ),
        ],
    ],
    "BADLY_FORMATTED_ROLE_EXO_LINK_KEY_DATA_PARENT": [
        "../test_data/delayed_errors/bad_role_exo_link_key_element_parent.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BADLY_FORMATTED_ROLE_EXO_LINK_KEY_DATA,
                "bad_role_exo_link_key_element_parent.ccpn/ccpnv3/memops/Implementation/bad_role_exo_link_key_element_parent.xml",
                _dedent_all("""in the file bad_role_exo_link_key_element_parent.xml in the link GUIW.WindowStore with guid default_user_2024-02-24-15-54-35-583_00005
                                    the key element nmrProject doesnt have children"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="bad_role_exo_link_key_element_parent",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00005 GUIW.WindowStore [keys: {{'nmrProject': '_ccp_nmr_Nmr_NmrProject___default___'}}]": "583_00005 GUIW.WindowStore - exo link detected but is incorrectly defined in root [error]"
                },
            ),
            (
                _dedent_all("""in the file bad_role_exo_link_key_element_parent.xml in the link GUIW.WindowStore with guid default_user_2024-02-24-15-54-35-583_00005
                           the key element nmrProject doesnt have children"""),
                False,
            ),
        ],
    ],
    "BADLY_FORMATTED_ROLE_EXO_LINK_KEY_DATA_CHILD": [
        "../test_data/delayed_errors/bad_role_exo_link_key_element_children.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.BADLY_FORMATTED_ROLE_EXO_LINK_KEY_DATA,
                "bad_role_exo_link_key_element_children.ccpn/ccpnv3/memops/Implementation/bad_role_exo_link_key_element_children.xml",
                _dedent_all("""in the file bad_role_exo_link_key_element_children.xml in the link GUIW.WindowStore with guid default_user_2024-02-24-15-54-35-583_00005
                                    the key element nmrProject doesnt have children"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="bad_role_exo_link_key_element_children",
                sub_directory="delayed_errors",
            ),
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00005 GUIW.WindowStore [keys: {{'nmrProject': '_ccp_nmr_Nmr_NmrProject___default___'}}]": "583_00005 GUIW.WindowStore - exo link detected but is incorrectly defined in root [error]"
                },
            ),
            (
                _dedent_all("""in the file bad_role_exo_link_key_element_children.xml in the link GUIW.WindowStore with guid default_user_2024-02-24-15-54-35-583_00005
                           the key element nmrProject doesnt have children"""),
                False,
            ),
        ],
    ],
    "FILE_NAME_HAS_NON_CCPN_ACSII_CHARACTERS": [
        "../test_data/delayed_errors/empty_good_project_with_non_ccpn_letter~.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NON_CCPN_ASCII_CHARACTER,
                "empty_good_project_with_non_ccpn_letter~.ccpn",
                _dedent_all("""the ccpn project directory name empty_good_project_with_non_ccpn_letter~.ccpn contains characters outside the set [a-zA-Z0-9_] which is not allowed
                         value: empty_good_project_with_non_ccpn_letter~.ccpn
                         ______________________________________________^_____
                      """),
            ),
            ErrorAndWarningData(
                ErrorCode.NON_CCPN_ASCII_CHARACTER,
                "empty_good_project_with_non_ccpn_letter~.xml",
                _dedent_all("""the ccpn root file name empty_good_project_with_non_ccpn_letter~.xml contains characters outside the set [a-zA-Z0-9_] which is not allowed
                      value: empty_good_project_with_non_ccpn_letter~.xml
                      ______________________________________________^____"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project_with_non_ccpn_letter~",
                sub_directory="warn_projects",
            ),
            (
                _dedent_all("""the ccpn project directory name empty_good_project_with_non_ccpn_letter~.ccpn contains characters outside the set [a-zA-Z0-9_] which is not allowed
                            value: empty_good_project_with_non_ccpn_letter~.ccpn
                            ______________________________________________^_____ [error]"""),
                False,
            ),
            (
                _dedent_all("""the ccpn root file name empty_good_project_with_non_ccpn_letter~.xml contains characters outside the set [a-zA-Z0-9_] which is not allowed
                           value: empty_good_project_with_non_ccpn_letter~.xml
                           ______________________________________________^____ [error]"""),
                False,
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
        ],
    ],
    "GOOD_PROJECT_EXTRA_ROOT_XML_FILES": [
        "../test_data/warn_projects/empty_good_project_extra_root_xml_files.ccpn",
        [
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project_extra_root_xml_files",
                sub_directory="warn_projects",
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS, sub_directory="warn_projects"),
            (
                _dedent_all("""more than one possible memops root file found in the implementation directory
                           empty_good_project_extra_root_xml_files.xml
                           empty_good_project_extra_root_xml_files_2.xml
                           [warning]"""),
                False,
            ),
            (
                "the file empty_good_project_extra_root_xml_files_2.xml is a possible orphaned memops root [warning]",
                False,
            ),
        ],
    ],
    "GOOD_PROJECT_MULTIPLE_ROOT_XML_FILES": [
        "../test_data/empty_good_project_multiple_root_xml_files.ccpn",
        [
            ErrorAndWarningData(
                code=ErrorCode.NO_MEMOPS_ROOT_FILES,
                cause="empty_good_project_multiple_root_xml_files.ccpn/ccpnv3/memops/Implementation",
                detail="""no xml files which could be memops roots were found in empty_good_project_multiple_root_xml_files.ccpn/ccpnv3/memops/Implementation
                          with the following extra messages being reported
                          more than one ccpn root file found in empty_good_project_multiple_root_xml_files.ccpn/ccpnv3/memops/Implementation and none match project name
                          roots are:
                          empty_good_project_multiple_root_xml_files.ccpn/ccpnv3/memops/Implementation/empty_good_project_mutiple_root_xml_files_2.xml
                          empty_good_project_multiple_root_xml_files.ccpn/ccpnv3/memops/Implementation/empty_good_project_mutiple_root_xml_files_1.xml""",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project_multiple_root_xml_files",
                discard_beyond="found an implementation directory",
            ),
            # *update_notes(EXPECTED_GOOD_EXO_LINKS),
            (
                _dedent_all("""more than one possible memops root file found in the implementation directory
                empty_good_project_mutiple_root_xml_files_2.xml
                empty_good_project_mutiple_root_xml_files_1.xml
                [warning]"""),
                False,
            ),
            (
                "the file empty_good_project_mutiple_root_xml_files_1.xml is a possible orphaned memops root [warning]",
                False,
            ),
            (
                "the file empty_good_project_mutiple_root_xml_files_2.xml is a possible orphaned memops root [warning]",
                False,
            ),
        ],
    ],
    "ROOT_EXO_LINK_MISSING_SOURCE_ELEMENT": [
        "../test_data/delayed_errors/root_exo_link_missing_source_element.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_LINK_ELEMENT_SOURCE_MISSING,
                "root_exo_link_missing_source_element.ccpn/ccpnv3/memops/Implementation/root_exo_link_missing_source_element.xml",
                _dedent_all("""in the file root_exo_link_missing_source_element.ccpn/ccpnv3/memops/Implementation/root_exo_link_missing_source_element.xml
                                    link MOLS.MolSystem with guid default_user_2024-02-24-15-54-35-583_00004 is missing its source element"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_exo_link_missing_source_element",
            ),
            # TODO: this uses pre template version boo!
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem [keys: {{'code': 'default'}}]": "583_00004 MOLS.MolSystem - exo link detected but is incorrectly defined in root [error]"
                },
            ),
            (
                _dedent_all("""in the file root_exo_link_missing_source_element.ccpn/ccpnv3/memops/Implementation/root_exo_link_missing_source_element.xml
                           link MOLS.MolSystem with guid default_user_2024-02-24-15-54-35-583_00004 is missing its source element [error]"""),
                False,
            ),
        ],
    ],
    "ROOT_EXO_LINK_TOO_MANY_SOURCE_ELEMENTS": [
        "../test_data/delayed_errors/root_exo_link_too_many_source_elements.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.EXO_LINK_ELEMENT_TOO_MANY_SOURCES,
                "root_exo_link_too_many_source_elements.ccpn/ccpnv3/memops/Implementation/root_exo_link_too_many_source_elements.xml",
                _dedent_all("""in the file root_exo_link_too_many_source_elements.ccpn/ccpnv3/memops/Implementation/root_exo_link_too_many_source_elements.xml
                                    link MOLS.MolSystem with guid default_user_2024-02-24-15-54-35-583_00004 has too many link elements [2] there should be 1"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_exo_link_too_many_source_elements",
            ),
            # TODO: this uses pre template version boo!
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00004 MOLS.MolSystem [keys: {{'code': 'default'}}]": "583_00004 MOLS.MolSystem - exo link detected but is incorrectly defined in root [error]"
                },
            ),
            (
                _dedent_all("""in the file root_exo_link_too_many_source_elements.ccpn/ccpnv3/memops/Implementation/root_exo_link_too_many_source_elements.xml
                           link MOLS.MolSystem with guid default_user_2024-02-24-15-54-35-583_00004 has too many link elements [2] there should be 1 [error]"""),
                False,
            ),
        ],
    ],
    "WARN_EMPTY_CONTAINERS": [
        "../test_data/warn_projects/empty_good_project_with_empty_containers.ccpn",
        [
            # ErrorData(ErrorCode.WARNING_EMPTY_CONTAINER,
            #           'empty_good_project_with_empty_containers.ccpn',
            #            dedent_all("""in the file root_exo_link_too_many_source_elements.ccpn/ccpnv3/memops/Implementation/root_exo_link_too_many_source_elements.xml
            #                         link MOLS.MolSystem with guid default_user_2024-02-24-15-54-35-583_00004 has too many link elements [2] there should be 1""")),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="empty_good_project_with_empty_containers",
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS, replace_text={}),
            (
                "empty directories [2] which may be orphaned containers found and listed below [warning]",
                False,
            ),
            (
                "  1. empty_good_project_with_empty_containers.ccpn/ccpnv3/ccp/molsim/Symmetry [warning]",
                True,
            ),
            (
                "  2. empty_good_project_with_empty_containers.ccpn/ccpnv3/ccp/molecule/LabeledMolecule [warning]",
                True,
            ),
        ],
    ],
    "EXO_LINKS_WITH_KEYS_OUTSIDE_THE_CCPN_CHARACTER_SET": [
        "../test_data/delayed_errors/exo_links_with_keys_outside_ccpn_character_set.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.NON_CCPN_ASCII_CHARACTER,
                ["default_user_2024-02-24-15-54-35-583_00006"],
                _dedent_all("""in the file exo_links_with_keys_outside_ccpn_character_set.ccpn/ccpnv3/memops/Implementation/exo_links_with_keys_outside_ccpn_character_set.xml
                               there are exo link keys [1] which contain characters outside the ccpn character set"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="exo_links_with_keys_outside_ccpn_character_set",
            ),
            # TODO: this uses pre template version boo!
            *update_notes(
                EXPECTED_GOOD_EXO_LINKS,
                replace_text={
                    "583_00006 GUIT.GuiTask [keys: {{'nameSpace': 'user', 'name': 'View'}}]":
                    "583_00006 GUIT.GuiTask [keys: {{'nameSpace': 'user', 'name': 'View~'}}]",
                    "00006 GUIT.GuiTask - [PROJECT] ccpnmr/gui/Task/user+View+default_user_2024-02-24-15-54-35-583_00006.xml":
                    "00006 GUIT.GuiTask - [PROJECT] ccpnmr/gui/Task/user+View~+default_user_2024-02-24-15-54-35-583_00006.xml"
                },
            ),
            (
                _dedent_all("""in the file exo_links_with_keys_outside_ccpn_character_set.ccpn/ccpnv3/memops/Implementation/exo_links_with_keys_outside_ccpn_character_set.xml
                         there are exo link keys [1] which contain characters outside the ccpn character set which are listed below [error]"""),
                False,
            ),
            (
                _dedent_all("""1. [name]. default_user_2024-02-24-15-54-35-583_00006 GUIT.GuiTask  name: View~ [error]
                              key: View~
                              _________^"""),
                False,
            ),
        ],
    ],
    "WARNING_ROOT_FILE_MISSING_TIME_ATTRIBUTE": [
        "../test_data/warn_projects/root_file_missing_time.ccpn",
        [
            # ErrorAndWarningData(
            #     ErrorCode.ROOT_FILE_TIME_ATTRIB_MISSING,
            #     'root_file_missing_time.ccpn/ccpnv3/ccp/Implementation/root_file_missing_time.xml',
            #     'in the file root_file_missing_time.xml the attribute time is missing in the element _StorageUnit'
            #     ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_file_missing_time",
                sub_directory="warn_projects",
                replace_text={
                    "memops root data was stored at Sat Feb 24 16:16:06 2024": "in the file root_file_missing_time.xml the attribute time is missing in the element _StorageUnit [warning]"
                },
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
        ],
    ],
    "WARNING_ROOT_FILE_TIME_ATTRIBUTE_BAD_FORMAT": [
        "../test_data/warn_projects/root_file_badly_formatted_time.ccpn",
        [
            # ErrorAndWarningData(
            #     ErrorCode.ROOT_FILE_TIME_ATTRIB_MISSING,
            #     'root_file_missing_time.ccpn/ccpnv3/ccp/Implementation/root_file_missing_time.xml',
            #     'in the file root_file_missing_time.xml the attribute time is missing in the element _StorageUnit'
            #     ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_file_badly_formatted_time",
                sub_directory="warn_projects",
                replace_text={
                    "memops root data was stored at Sat Feb 24 16:16:06 2024": "in the file root_file_badly_formatted_time.xml the attribute time [2024 Sat Feb 24 16:16:06] is badly formatted in the element _StorageUnit [warning]"
                },
            ),
            *update_notes(EXPECTED_GOOD_EXO_LINKS),
        ],
    ],
    "ROOT_VERSION_MISSING": [
        "../test_data/root_version_missing.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.ROOT_MODEL_VERSION_MISSING,
                "root_version_missing.ccpn/ccpnv3/memops/Implementation/root_version_missing.xml",
                "in the file root_version_missing.xml the attribute release is missing in the element _StorageUnit",
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_version_missing",
                sub_directory="warn_projects",
                discard_beyond="ccpn project memops root file found in",
            ),
        ],
    ],
    "ROOT_VERSION_BAD_FORMAT": [
        "../test_data/root_version_badly_formatted.ccpn",
        [
            ErrorAndWarningData(
                ErrorCode.ROOT_MODEL_VERSION_BAD,
                "root_version_badly_formatted.ccpn/ccpnv3/memops/Implementation/root_version_badly_formatted.xml",
                _dedent_all("""in the file root_version_badly_formatted.xml the attribute release [a3.1.0]
                              is not a valid version number it should be of the form <major>.<minor>.<patch> where <major>,
                              <minor> can only contain digits and <patch> can contain digits with a possible appended letter
                              can't continue as can't determine model version"""),
            ),
            *update_notes(
                EXPECTED_GOOD_NOTES_READ_MEMOPS_ROOT,
                file_name="root_version_badly_formatted",
                sub_directory="warn_projects",
                discard_beyond="ccpn project memops root file found",
            ),
        ],
    ],
}

TEST_IDS = list(expecteds.keys())

ERROR_CODES_READ_PROTECTED = [
    error_code for error_code in TEST_IDS if "READ_PROTECTED" in error_code
]
ERROR_CODES_NOT_READ_PROTECTED = [
    error_code for error_code in TEST_IDS if "READ_PROTECTED" not in error_code
]
FILES_TO_PROTECT_BY_ERROR_CODE = {
    "READ_PROTECTED_PROJECT": [
        ("no_read_empty_project.ccpn", False),
    ],
    "READ_PROTECTED_CCPNV3": [
        ("no_read_ccpnv3.ccpn/ccpnv3", False),
    ],
    "MEMOPS_ROOT_READ_PROTECTED_GOOD_PROJECT": [
        (
            "memops_root_no_read_empty_good_project.ccpn/ccpnv3/memops/Implementation/memops_root_no_read_empty_good_project.xml",
            False,
        ),
    ],
}
ERROR_CODES_TO_READ_PROTECTED_FILES = {
    ErrorCode.READ_PROTECTED_PROJECT: ("no_read_empty_project.ccpn", False),
    ErrorCode.READ_PROTECTED_CCPNV3: ("no_read_ccpnv3.ccpn/ccpnv3", False),
}

ERRORS_NOT_NEEDING_TEST_SETS = set([ErrorCode.INTERNAL_ERROR, ErrorCode.MISSING_ATTRIB])



import json

from os import walk
from pathlib import Path
import xml.etree.ElementTree as et
import sys
from textwrap import dedent

from project_checker.DiskModelChecker import ObjectInfo

top_object_info_map = {}
guid_to_type = {}
short_name_to_guid = {}

TOP_OBJECT_GUID = "www.ccpn.ac.uk_Fogh_2006-09-14-16:28:57_00002"
# key types currently not in use
# LINE_GUID = "www.ccpn.ac.uk_Fogh_2006-08-16-14:22:53_00033"
# WORD_GUID= "www.ccpn.ac.uk_Fogh_2006-08-16-14:22:53_00037"
# MOL_SYSTEM = 'www.ccpn.ac.uk_Fogh_2006-08-16-14:22:54_00023'
# LONG_WORD = 'www.ccpn.ac.uk_Fogh_2007-09-12-18:31:28_00003'
# MOL_TYPE = 'www.ccpn.ac.uk_Fogh_2006-08-16-14:22:52_00024'
# INT = 'www.ccpn.ac.uk_Fogh_2006-08-16-14:22:53_00032'


def _get_parents_upto(file_path, root_name):

    parts = list(reversed(file_path.parts))
    xml_index = parts.index(root_name)

    return list(reversed(parts[1:xml_index]))


def _get_value_type(key_object):
    return [elem.text for elem in key_object if elem.tag == "valueType"][0]


def _analyse_file(root, file_path):
    result = None

    guid = root.attrib["guid"]

    parent_guid = root.attrib["container"] if "container" in root.attrib else None
    name = root.attrib["name"]
    super_types = [elem for elem in root if elem.tag == "supertypes"]
    location = _get_parents_upto(file_path, "xml")

    # really we should be following the package hierarchy to get the containment
    if root.tag == "MetaPackage":
        if name == "Root":
            guid_to_type[guid] = [
                name,
            ]
        else:
            guid_to_type[guid] = location

        short_name = root.attrib["shortName"] if "shortName" in root.attrib else None
        if short_name:
            short_name_to_guid[short_name] = guid
    else:
        guid_to_type[guid] = [*location, name]

    if super_types:
        super_type_guids = [elem.text for elem in super_types[0] if elem.tag == "item"]

        keys = _get_key_names(root)

        key_objects = [
            elem
            for elem in root
            if "name" in elem.attrib and elem.attrib["name"] in keys
        ]

        key_model_types = {
            key_object.attrib["name"]: key_object.tag for key_object in key_objects
        }

        key_value_types = {
            key_object.attrib["name"]: _get_value_type(key_object)
            for key_object in key_objects
        }

        default_items = _get_key_defaults(key_objects)

        result = ObjectInfo(
            name=name,
            guid=guid,
            supertype_guids=super_type_guids,
            containment=location,
            parent_guid=parent_guid,
            keys=keys,
            key_type_guids=key_value_types,
            key_model_types=key_model_types,
            key_defaults=default_items,
        )

    return result


def _get_key_defaults(key_objects):
    result = {}
    for key_object in key_objects:
        name = key_object.attrib["name"]

        default_value_elements = [
            elem for elem in key_object if elem.tag == "defaultValue"
        ]
        if len(default_value_elements) > 0:
            result[name] = [elem.text for elem in default_value_elements[0]][0]

    return result


def _get_key_names(root):
    key_names = [elem for elem in root if elem.tag == "keyNames"]
    if key_names:
        keys = [elem.text for elem in key_names[0] if elem.tag == "item"]
    else:
        keys = []
    return keys


def _load_file(file_path):
    with open(file_path, "r") as fh:
        root = None
        try:
            root = et.fromstring(fh.read())
        except et.ParseError as e:
            msg = f"""
            Error: the file {file_path.parts[-1]} can't be parsed as XML
            ****************************************************************
            error code: XML_PARSE_ERROR
            caused by: {file_path}
            detailed message:
            {e}
            ****************************************************************"""
            print(dedent(msg), file=sys.stderr)

    return root


def _patch_tree(tree):
    # PairwiseConstraintItems don't have keys... they should!
    if tree.attrib["guid"] == "www.ccpn.ac.uk_Fogh_2007-10-17-10:43:15_00001":
        print(
            "Note: modifying PairwiseConstraintItem to have KeyNames", file=sys.stderr
        )
        key_names = et.Element("keyNames")
        tree.append(key_names)
        resonances_item = et.Element("item")
        resonances_item.text = "resonances"
        key_names.append(resonances_item)

    return tree


def _walk_meta_model(directory_path):
    print(f"starting walk of model xml files from {directory_path}")

    all_top_object_info = {}
    for directory_path, _, file_names in walk(directory_path):
        for file_name in file_names:
            file_path = Path(directory_path) / file_name
            if file_path.suffix != ".xml":
                continue
            tree = _load_file(file_path)
            top_object_info = None
            if tree:
                tree = _patch_tree(tree)
                top_object_info = _analyse_file(tree, file_path)
            if top_object_info:
                all_top_object_info[top_object_info.guid] = top_object_info

    return all_top_object_info


def _find_all_super_types(top_object_info, top_object_info_map):
    super_type_guids = top_object_info.supertype_guids
    super_type_objects = [
        top_object_info_map[guid]
        for guid in super_type_guids
        if guid in top_object_info_map
    ]
    for super_type_object_info in super_type_objects:
        parent_objects = _find_all_super_types(
            super_type_object_info, top_object_info_map
        )
        super_type_objects.extend(parent_objects)

    super_type_by_guid = {object.guid: object for object in super_type_objects}

    return super_type_by_guid.values()


# from https://stackoverflow.com/questions/3494906/how-do-i-merge-a-list-of-dicts-into-a-single-dict
def _merge_dicts(dict_list):
    return {key: value for child_dict in dict_list for key, value in child_dict.items()}


def _find_values_over_hierarchy(top_object_info, field_name, top_object_info_map):
    top_object_info_list = [
        top_object_info,
        *_find_all_super_types(top_object_info, top_object_info_map),
    ]
    key_types = [
        getattr(top_object_info, field_name) for top_object_info in top_object_info_list
    ]
    return _merge_dicts(key_types)


def build_top_info(model_version):
    global top_object_info_map
    root = Path(__file__).parent.parent
    model_root = root / "ccpnmodel" / "versions"
    model_version_root = model_root / model_version
    top_object_info_map = _walk_meta_model(model_version_root)

    for top_object_info in top_object_info_map.values():
        top_object_info.supertype_names = [
            guid_to_type[guid] for guid in top_object_info.supertype_guids
        ]
        top_object_info.key_types_names = {
            key: guid_to_type[guid]
            for key, guid in top_object_info.key_type_guids.items()
        }

        key_defaults = _find_values_over_hierarchy(
            top_object_info, "key_defaults", top_object_info_map
        )
        key_type_guids = _find_values_over_hierarchy(
            top_object_info, "key_type_guids", top_object_info_map
        )
        key_model_types = _find_values_over_hierarchy(
            top_object_info, "key_model_types", top_object_info_map
        )
        key_types_names = _find_values_over_hierarchy(
            top_object_info, "key_types_names", top_object_info_map
        )

        top_object_info.key_defaults = key_defaults
        top_object_info.key_type_guids = key_type_guids
        top_object_info.key_model_types = key_model_types
        top_object_info.key_types_names = key_types_names


if __name__ == "__main__":
    import time

    start = time.time()

    model_version = "v_3_1_0"
    build_top_info(model_version)

    end = time.time()

    print(f"elapsed time: {end - start}")

    file_path = Path(__file__)
    model_info_path = file_path.parent / "model_info"
    with open(model_info_path / f"{model_version}_object_info.json", "w") as fh:
        json_data = json.dumps(
            top_object_info_map, default=lambda obj: obj.__dict__, indent=4
        )  # json.dump(top_object_info[0], fh,)
        fh.write(json_data)

    with open(
        model_info_path / f"{model_version}_guid_to_storage_location.json", "w"
    ) as fh:
        json.dump(guid_to_type, fh, indent=4)

    with open(model_info_path / f"{model_version}_short_name_to_guid.json", "w") as fh:
        json.dump(short_name_to_guid, fh, indent=4)



from .nothing import Nothing
from .something import Something


class Optional(object):

    def get(self): ...

    @classmethod
    def of(cls, thing=None, messages=(), error_code=None, cause=None):
        if (thing and messages) or (thing and error_code) or (thing and cause):
            NEWLINE = "\n"
            msg = f"""
            got messages / error / cause  code but thing is not None
            thing 
            -----
            {thing}
            
            messages
            --------
            {NEWLINE.join(messages)}
            
            error code
            ----------
            
            {error_code.value}
            """
            raise Exception(msg)

        result = (
            Something(thing, cls)
            if thing
            else Nothing(cls, messages=messages, error_code=error_code, cause=cause)
        )
        return result

    @classmethod
    def empty(cls, messages=(), error_code=None, cause=None):
        return Nothing(cls, messages=messages, error_code=error_code, cause=cause)



# https://stackoverflow.com/questions/75048986/way-to-temporarily-change-the-directory-in-python-to-execute-code-without-affect
import contextlib
import os
import shutil
import stat
import sys
from difflib import SequenceMatcher
from io import StringIO
from textwrap import dedent
from time import sleep

from pathlib import Path

from project_checker.tests.api_test_data import ERROR_CODES_NOT_READ_PROTECTED, expecteds

TEST_DIRECTORY = Path(__file__).parent / '..' / 'test_data'
TEST_FILE = TEST_DIRECTORY / 'cli_test_output.txt'

class RedirectedStdoutToMemory:
    def __init__(self):
        self.new_output = StringIO()

    def __enter__(self):
        self.saved_stdout_output = sys.stdout
        self.saved_stderr_output = sys.stderr
        sys.stdout = self.new_output
        sys.stderr = self.new_output

        return self.new_output

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.saved_stdout_output
        sys.stderr = self.saved_stderr_output


from project_checker.DiskModelChecker import (
    ErrorAndWarningData,
    ModelChecker,
    ExitStatus, run_cli_checker,
)
from project_checker.optional import Optional


def check_results_against_expected(expected, file_path, test_directory):
    expected = list(expected)

    with different_cwd(test_directory):
        checker = ModelChecker()
        project_root = Path(test_directory).parts[-1]

        good_project = project_root == "good_projects"
        warn_project = project_root == "warn_projects"
        delayed_error_project = project_root == "delayed_errors"

        result = checker.run(file_path)

        if good_project:
            expected_exit_status = ExitStatus.EXIT_OK
        elif warn_project:
            expected_exit_status = ExitStatus.EXIT_WARN
        elif delayed_error_project:
            expected_exit_status = ExitStatus.EXIT_ERROR
        else:
            expected_exit_status = ExitStatus.EXIT_ERROR_INCOMPLETE

        if result != expected_exit_status:
            print(f"expected exit status {expected_exit_status} but got {result}")

        missing_errors = []
        for error in checker.errors:
            if isinstance(error, ErrorAndWarningData):
                found_error = assert_error_in_expected_and_return(error, expected)
                if not found_error:
                    msg = f"""\
                        failed to find error
                        code: {error.code}
                        cause: {error.cause}
                        detail: {error.detail}

                        because {found_error.messages}
                    """
                    print(dedent(msg), file=sys.stderr)
                    missing_errors.append(error)
                else:
                    found_error = found_error.get()
                    expected.remove(found_error)

        unexpected_messages = []
        for message in checker.messages:
            if message[0] == "":
                continue

            if 'in stand alone mode' in message[0]:
                continue

            if message in expected:
                # print('unexpected message:', message)
                expected.remove(message)
            else:
                unexpected_messages.append(message)

        BAR = "|"
        if unexpected_messages:
            unexpected_messages = [
                f"{BAR}{message}{BAR}" for message in unexpected_messages
            ]

            print("unexpected messages:")
            for unexpected_message in unexpected_messages:
                print(dedent(unexpected_message), file=sys.stderr)

        remaining_messages = [
            f"{BAR}{message}{BAR}"
            for message in expected
            if not isinstance(message, ErrorAndWarningData)
        ]
        if remaining_messages:
            print("remaining messages:", file=sys.stderr)
            for remaining_message in remaining_messages:
                print(dedent(remaining_message), file=sys.stderr)

            print("checker errors:")
            for error in checker.errors:
                print(dedent(str(error)), file=sys.stderr)
        
        if len(expected) > 0:
            print("expected messages and which were not found:", expected)

        assert missing_errors == []

        assert result == expected_exit_status

        assert unexpected_messages == []

        assert expected == []


@contextlib.contextmanager
def different_cwd(x):
    d = os.getcwd()

    # This could raise an exception, but it's probably
    # best to let it propagate and let the caller
    # deal with it, since they requested x
    os.chdir(x)

    try:
        yield

    finally:
        # This could also raise an exception, but you *really*
        # aren't equipped to figure out what went wrong if the
        # old working directory can't be restored.
        os.chdir(d)


def get_test_project_and_working_directory(file_path):
    file_path = Path(file_path)
    project_index = None

    for i, part in enumerate(list(file_path.parts)):
        if part.endswith(".ccpn") or part.endswith("_ccpn"):
            project_index = i
            break

    if not project_index:
        raise ValueError(f"couldn't find project directory in {file_path}")

    directory_offset = project_index
    project_path = Path(*file_path.parts[directory_offset:])
    working_directory = Path(*file_path.parts[:directory_offset])
    return project_path, working_directory


def make_path_unreadable_by_owner(path):
    old_mode = stat.S_IMODE(os.lstat(path).st_mode)
    new_mode = old_mode & ~stat.S_IRUSR
    os.chmod(path, new_mode)


def path_is_readable_by_owner(path):
    current_permissions = stat.S_IMODE(os.lstat(path).st_mode)

    return bool(current_permissions & stat.S_IRUSR)


def setup_read_protected(file_path, tmp_path, files_to_protect=()):
    file_path = Path(file_path).resolve()
    tmp_path = Path(tmp_path)

    new_root = tmp_path / file_path.parts[-1]
    shutil.copytree(file_path, new_root)

    can_read_path = False
    for i in range(10):
        for target, recursive in files_to_protect:
            target_path = tmp_path / target
            make_path_unreadable_by_owner(target_path)
            can_read_path = path_is_readable_by_owner(target_path)
            if can_read_path == 0:
                sleep(0.01)
            else:
                break

    assert (
        not bool(can_read_path)
    ), f"couldn't set the path {target_path} to read only!"

    return new_root


def match_ratio(expected, actual):
    matcher = SequenceMatcher(None, expected, actual)
    return matcher.ratio()


def lstrip_all(text):
    return "\n".join([elem.lstrip() for elem in text.split("\n")])


def assert_error_in_expected_and_return(error, expected, match_distance=0.99):
    result = None
    failures = []

    error_count = 0
    for expected_error in expected:
        if isinstance(expected_error, ErrorAndWarningData):
            error_count += 1

            # for some reason pytest? is screwing up the comparison of the error codes possibly by reloading the enum
            # so we have to compare the values rather than the ids of the enums
            if error.code.value != expected_error.code.value:
                failures.append(
                    f"error code expected: {expected_error.code}, found: {error.code}"
                )
                continue

            if str(expected_error.cause) != str(error.cause):
                failures.append(
                    f"cause expected: {expected_error.cause[:30]}..., found: {error.cause[:30]}..."
                )
                continue

            detail_distance = SequenceMatcher(
                None, lstrip_all(expected_error.detail), lstrip_all(error.detail)
            ).ratio()
            if detail_distance < match_distance:
                failures.append(
                    f"detail expected: {expected_error.detail[:30]}..., found: {error.detail[:30]}..."
                )
                continue

            result = expected_error
            break

    if not result and not error_count:
        failures = ["no errors defined!"]

    return Optional.of(result, messages=failures)



def build_cli_output():
    with RedirectedStdoutToMemory() as saved_stdout:
        print(file=sys.stderr)
        max_length_of_test_case_name = max(
            [len(test_case) for test_case in ERROR_CODES_NOT_READ_PROTECTED]
        )
        for test_case in ERROR_CODES_NOT_READ_PROTECTED:
            print(file=sys.stderr)
            test_case_name = f" {test_case} ".center(max_length_of_test_case_name, "-")
            print(f"{'-'*20}{test_case_name}{'-'*20}", file=sys.stderr)
            file_path, expected = expecteds[test_case]

            tests_directory = Path(__file__).parent

            file_path = tests_directory / file_path

            project_path, working_directory = get_test_project_and_working_directory(
                file_path
            )

            with different_cwd(working_directory):
                exit_code = run_cli_checker(project_path)

            print(file=sys.stderr)
            print(
                f"command exited with exit code: {exit_code}, [{ExitStatus(exit_code)}]",
                file=sys.stderr,
            )

        print(file=sys.stderr)
        dummy_name = '-'*max_length_of_test_case_name
        filler = f'-{dummy_name}-'
        print(f"{'-' * 20}{filler}{'-' * 20}", file=sys.stderr)

    return saved_stdout.getvalue()



from project_checker.tests.testing_utils import build_cli_output, TEST_FILE
from freezegun import freeze_time

@freeze_time("2012-01-14")
def make_cli_test_file():

    test_output = build_cli_output()
    with open(TEST_FILE, 'w') as f:
        f.write(test_output)


if __name__ == '__main__':
    make_cli_test_file()


"""compatible_abc

This module exports a single class, CompatibleABC.
It is necessary to provide the same behavior in
Python 2 and Python 3.

The implementation was taken from https://stackoverflow.com/a/38668373
"""

from abc import ABCMeta


CompatibleABC = ABCMeta("ABC", (object,), {"__slots__": ()})



from .abstract_optional import AbstractOptional
from .exceptions import OptionalAccessOfEmptyException


class Nothing(AbstractOptional):
    def __init__(self, optional, target=None, messages=(), error_code=None, cause=None):
        self.__optional = optional
        self.target = target
        self.messages = tuple(messages)
        self.error_code = error_code
        self.cause = cause

    def is_empty(self):
        return True

    def get(self):
        raise OptionalAccessOfEmptyException("You cannot call get on an empty optional")

    def get_or_default(self, default_value):
        return default_value

    def get_or_raise(self, raiseable):
        raise raiseable

    def if_present(self, consumer):
        return self

    def or_else(self, supplier):
        return self.__optional.of(supplier())

    def or_else_raise(self, raiseable):
        raise raiseable

    def map(self, func):
        return self

    def flat_map(self, func):
        return self

    def __eq__(self, other):
        return isinstance(other, Nothing)

    def __repr__(self):
        return f"Optional.empty(messages={self.messages}"



from difflib import SequenceMatcher


from project_checker.tests.testing_utils import  TEST_FILE, build_cli_output

def test_cli_checker(time_machine):
    time_machine.move_to(0, tick=False)
    EXPECTED = open(TEST_FILE).read().split('\n')
    EXPECTED_SET = set(EXPECTED)
    test_output = build_cli_output()
    for i,line in enumerate(test_output.splitlines()):
        if i <= 1:
            continue

        if 'in stand alone mode' in line:
            continue

        if not line in EXPECTED_SET:
            best_matches = {}
            matcher = SequenceMatcher()
            for j, expd in enumerate(EXPECTED):
                matcher.set_seqs(line, expd)
                best_matches.setdefault(matcher.ratio(),[]).append(((i, line), (j,expd)))
                best_match =  best_matches[max(best_matches.keys())]

                for match in best_match:
                    print(f"line: {match[0][0]} {match[0][1]}")
                    print(f"expd: {match[1][0]} {match[1][1]}")

                assert False, f"line not found: {line} in expected"




